use glam::uvec2;
use hui::{
  draw::{ImageHandle, TextureFormat},
  element::{container::Container, image::Image, UiElementExt},
  layout::Alignment,
  size
};
use shipyard::{AllStoragesViewMut, IntoIter, NonSendSync, Unique, UniqueView, UniqueViewMut, View};
use crate::{hui_integration::UiState, player::MainPlayer, rendering::Renderer, settings::GameSettings, world::raycast::LookingAtBlock};

const CROSSHAIR_SIZE: usize = 9;
const CROSSHAIR: &[u8] = &[
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
];
const CROSSHAIR_ALT: &[u8] = &[
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
  0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
  0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
  0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
  0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
  0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
  0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
];

#[derive(Unique)]
pub struct CrosshairImage(ImageHandle, ImageHandle);

pub fn init_crosshair_image(storages: AllStoragesViewMut) {
  let mut ui = storages.borrow::<NonSendSync<UniqueViewMut<UiState>>>().unwrap();
  let image = ui.hui.add_image(TextureFormat::Grayscale, CROSSHAIR, CROSSHAIR_SIZE);
  let image_alt = ui.hui.add_image(TextureFormat::Grayscale, CROSSHAIR_ALT, CROSSHAIR_SIZE);
  storages.add_unique(CrosshairImage(image, image_alt));
}

pub fn draw_crosshair(
  mut ui: NonSendSync<UniqueViewMut<UiState>>,
  crosshair: UniqueView<CrosshairImage>,
  ren: UniqueView<Renderer>,
  player: View<MainPlayer>,
  raycast: View<LookingAtBlock>,
  settings: UniqueView<GameSettings>,
) {
  let mut active = !settings.dynamic_crosshair;
  if settings.dynamic_crosshair {
    if let Some((_, raycast)) = (&player, &raycast).iter().next() {
      active = raycast.0.is_some();
    }
  }

  let size = ren.size_uvec2();
  let size_rounded = uvec2(size.x & !1, size.y & !1).as_vec2();

  Container::default()
    .with_size(size!(100%))
    .with_align(Alignment::Center)
    .with_children(|ui| {
      Image::new(if active { crosshair.0 } else { crosshair.1 })
        .with_color((1., 1., 1., 0.5))
        .with_size(size!((CROSSHAIR_SIZE * 2)))
        .add_child(ui);
    })
    .add_root(&mut ui.hui, size_rounded);
}
